#!/usr/bin/env python
"""
CLI for controlling mongodb

Author: Jacob Feder
Date: 9/13/2020
"""
import argparse
from sys import platform, exit
from pathlib import Path
from subprocess import check_call, CalledProcessError

from nspyre.errors import OSNotSupportedError


THIS_DIR = Path(__file__).parent


def main():
    """Entry point for mongodb CLI"""
    # parse command-line arguments
    arg_parser = argparse.ArgumentParser(prog='nspyre-mongodb', description='Start / restart the MongoDB server')
    cmd_args = arg_parser.parse_args()
    try:
        if platform == 'linux' or platform == 'linux2' or platform == 'darwin':
            check_call(['bash', str(THIS_DIR / 'start_mongo_unix.sh')])
        elif platform == 'win32':
            check_call([str(THIS_DIR / 'start_mongo_win.bat')])
        else:
            raise OSNotSupportedError('Your OS [{}] is not supported'.\
                                        format(platform))
    except CalledProcessError:
        exit(1)

    exit(0)


if __name__ == '__main__':
    main()



"""The class defining the NSpyre Instrument Manager MainWindow.

The InstrumentManagerWindow class is the main GUI window for viewing the live
state and settings of the hardware devices connected to an NSpyre InstrumentServer.
It is defined by the QtWidgets.QMainWindow subclass from Qt and consists of a
QTreeWidget and subsequent QTreeWidgetItem(s) for displaying the attributes of
each device located on each connected Instrument Server. It is also responsible
for creating and connecting all the callback functions to the device drivers'
PySignal.Signal(s) for updating it's own GUI QtWidgets.QtWidget(s) in real-time.
i.e. QComboBox, QLineEdit, SpinBox, etc.

From the Qt documentation:
The QMainWindow class provides a main application window, with a menu bar, dock
windows (e.g. for toolbars), and a status bar. Main windows are most often used to
provide menus, toolbars and a status bar around a large central widget, such as a
text edit, drawing canvas or QWorkspace (for MDI applications). QMainWindow is
usually subclassed since this makes it easier to encapsulate the central widget,
menus and toolbars as well as the window's state. Subclassing makes it possible to
create the slots that are called when the user clicks menu items or toolbar buttons.

  Typical usage example:

  app = app.NSpyreApp([sys.argv])
  pyqtgraph._connectCleanup()
  with ..inserv.gateway.InservGateway() as isg:
      window = main_window.InstrumentManagerWindow(isg)
      sys.exit(app.exec())

Copyright (c) 2020, Alexandre Bourassa, Michael Solomon, Jacob Feder
All rights reserved.

This work is licensed under the terms of the 3-Clause BSD license.
For a copy, see <https://opensource.org/licenses/BSD-3-Clause>.
"""
import functools
import logging

from pimpmyclass.helpers import DictPropertyNameKey
from PyQt5.QtCore import QSize
from PyQt5.QtGui import QColor, QFont
from PyQt5.QtWidgets import QApplication, QComboBox, QHeaderView, QLineEdit, QMainWindow, QPushButton, QTreeWidget, QTreeWidgetItem
from pyqtgraph import SpinBox as pyqtgraph_SpinBox
from pyqtgraph import _connectCleanup as pyqtgraph_connectCleanup
from pint.util import infer_base_unit

from nspyre.config.config_files import load_meta_config
from nspyre.definitions import Q_, CLIENT_META_CONFIG_PATH
from nspyre.errors import InstrumentManagerError
from nspyre.inserv.gateway import InservGateway

logger = logging.getLogger(__name__)


class InstrumentManagerWindow(QMainWindow):
    """Creates a GUI interface for controlling instrument attributes
    based on a dropdown / tree structure
    the top level is the instrument servers
    the next level is the devices
    the bottom level is attributes of the devices.

     # tree of dictionaries that contains all of the instrument
     # manager GUI elements
     # the top level of the dictionary is the instrument servers
     # the next level is the devices
     # the bottom level is attributes of the devices

     e.g.               ----------self.gui----------
                       /                            \
                   server1                        server2
                  /       \                      /       \
          sig-gen1         scope1        sig-gen2         laser
          /      \        /      \      /        \       /     \
       freq     ampl    trig    din[] freq      ampl  lambda  power
     """

    def __init__(self, gateway, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('NSpyre Instrument Manager')
        # connection to the instrument servers
        self.gateway = gateway

        # set main GUI layout
        self.tree = QTreeWidget()
        self.tree.setFont(QFont('Helvetica [Cronyx]', 14))
        self.tree.setColumnCount(2)
        self.tree.setMinimumHeight(self.tree.height())
        # self.tree.setUniformRowHeights(True)
        # self.tree.setHeaderLabels(['Lantz Feat', 'value'])
        # self.tree.setDragEnabled(True)
        # self.tree.setSortingEnabled(True)
        # self.tree.sortByColumn(0, Qt.AscendingOrder)

        # configure the QTreeWidget Header
        header = self.tree.header()
        header.setHidden(True)
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        # header.setSectionResizeMode(0, QHeaderView.Fixed) QHeaderView.Stretch, QHeaderView.Stretch, QHeaderView.Interactive
        header.setSectionsMovable(True)
        header.setStretchLastSection(True)  # False

        # generate gui elements and set minimum window dimensions before displaying
        self._create_widgets()
        self.setCentralWidget(self.tree)
        self.tree.expandAll()
        # Adding 2pt of padding to the margin to remove horizontal scroll bar and 14pt of padding for the vertical scroll bar
        self.tree.setMinimumWidth(self.tree.columnWidth(0) + self.tree.columnWidth(1) + 2 + 14)
        self.tree.collapseAll()
        # start GUI with the servers expanded
        for i in range(self.tree.topLevelItemCount()):
            self.tree.topLevelItem(i).setExpanded(True)
        self.show()


    def _create_widgets(self):
        """Iterate over the available servers and devices, collect their
        attributes that can be modified by the instrument manager GUI, then
        populate the self.gui"""

        # iterate over servers
        for server_name, server in self.gateway.servers().items():
            server_tree = QTreeWidgetItem(self.tree, [server_name, ''])
            server_tree.setExpanded(True)

            # iterate over devices
            for device_name, device in server.root._devs.items():
                device_tree = QTreeWidgetItem(server_tree, [device_name, ''])
                device_tree.setExpanded(True)

                try:
                    # handle feats
                    # TODO this also returns dictfeats!
                    for feat_name, feat in device._lantz_feats.items():
                        # filter out any dictfeats
                        if isinstance(feat_name, DictPropertyNameKey):
                            continue
                        feat_widget, feat_getattr_func = self._generate_feat_widget(feat, feat_name, device)
                        # we have to use a partial here because PySignal and RPyC don't
                        # play nicely if you .connect() a lambda or other function / method
                        # to PySignal
                        feat_getattr_func.__name__ = 'InstrumentManager_getattr_func'
                        getattr_partial = functools.partial(feat_getattr_func)
                        getattr_partial = functools.update_wrapper(getattr_partial, feat_getattr_func)
                        # register the feat_getattr_func() to be called when the feat changes using
                        # pimpmyclass "ObservableProperty" mixin
                        getattr(device, feat_name + '_changed').connect(getattr_partial)

                        feat_widget.setFont(QFont('Helvetica [Cronyx]', 14))
                        feat_item = QTreeWidgetItem(device_tree, [feat_name, ''])
                        feat_item.sizeHint(0)
                        feat_item.setSizeHint(1, QSize(79,24))
                        self.tree.setItemWidget(feat_item, 1, feat_widget)

                    # handle dictfeats
                    for dictfeat_name, dictfeat in device._lantz_dictfeats.items():
                        """Generate a Qt gui element for a lantz dictfeat"""
                        dictfeat_tree = QTreeWidgetItem(device_tree, [dictfeat_name, ''])

                        # TODO
                        # dummy 'get' of dict feat value in order to force lantz to populate
                        # its 'subproperties' TODO this is pretty hacky
                        for feat_key in dictfeat.keys:
                            feat = dictfeat.subproperty(getattr(device, dictfeat_name).instance, feat_key)
                            feat_widget, feat_getattr_func = self._generate_feat_widget(feat, dictfeat_name, device, dictfeat_key=feat_key)
                            feat_item = QTreeWidgetItem(dictfeat_tree, ['{} {}'.format(dictfeat_name, feat_key), ''])
                            self.tree.setItemWidget(feat_item, 1, feat_widget)
                            if feat_key == dictfeat.keys[-1]:
                                # connect the feat_getattr() to be called when the feat changes using
                                # pimpmyclass "ObservableProperty" mixin
                                def dictfeat_getattr_func(df_tree, df_keys, getattr_func, value, old_value, key):
                                    w = self.tree.itemWidget(df_tree.child(df_keys.index(key)), 1)
                                    getattr_func(value, old_value, widget=w)

                                # we have to use a partial here because PySignal and RPyC don't
                                # play nicely if you .connect() a lambda or other function / method
                                # to PySignal
                                feat_getattr_func.__name__ = 'InstrumentManager_getattr_func'
                                getattr_partial = functools.partial(dictfeat_getattr_func, dictfeat_tree, dictfeat.keys, feat_getattr_func)
                                getattr_partial = functools.update_wrapper(getattr_partial, feat_getattr_func)
                                getattr(device, dictfeat_name + '_changed').connect(getattr_partial)

                    # handle actions
                    action_tree = QTreeWidgetItem(device_tree, ['Actions', ''])
                    for action_name, action in device._lantz_actions.items():
                        # actions that shouldn't be added to the GUI
                        ignore_actions = ['initialize', 'finalize', 'update', 'refresh']
                        if action_name in ignore_actions or '_async' in action_name:
                            continue

                        action_widget = self._generate_action_widget(device, action, action_name)
                        action_widget.setFont(QFont('Helvetica [Cronyx]', 14))
                        action_item = QTreeWidgetItem(action_tree, [action_name, ''])
                        self.tree.setItemWidget(action_item, 1, action_widget)
                except Exception as exc:
                    logger.error(exc)
                    # some error has occured loading the device attributes
                    device_tree.setText(0, 'ERROR: {}'.format(device_name))
                    device_tree.setBackground(0, QColor(255, 0, 0, 150))
                    pass


    def _generate_feat_widget(self, feat, feat_name, device, dictfeat_key=None):
        """Generate a Qt GUI element for a lantz feat/dictfeat"""

        if dictfeat_key:
            ## val = dictfeat.subproperty(getattr(device, feat_name).instance, dictfeat_key)
            # val = getattr(device, feat_name)[dictfeat_key]  # .__getitem__(dictfeat_key)
            feat_value = getattr(device, feat_name)[dictfeat_key]
            # if lantz has a function pointer in df.fset, then it is writeable
            read_only = False if getattr(device, feat_name).df.fset else True
        else:
            feat_value = getattr(device, feat_name)
            read_only = False if feat.fset else True
        
        # the lantz feat has only a specific set of allowed values
        # so we make a dropdown box
        if feat._config['values']:
            if read_only:
                # use a lineedit instead of a combo box if it's read only
                widget = QLineEdit()
                widget.setText(str(feat_value))
                widget.setReadOnly(True)
                def getattr_func(value, old_value, widget=widget):
                    widget.setText(value)
            else:
                widget = QComboBox()
                # print('combo box' + str(widget.sizeHint()))
                # dictionary mapping the possible lantz values to str(values)
                # e.g. {'True' : True, 'False' : False}
                keymapping_dict = {}
                for k in iter(feat._config['values']):
                    keymapping_dict[str(k)] = k

                # add the possible values to the dropdown list
                widget.addItems(keymapping_dict.keys())
                widget.setCurrentIndex(list(keymapping_dict.values()).index(feat_value))

                # callback function to modify the GUI when the the feat is changed
                # externally
                # we have to use a partial here because PySignal and RPyC don't
                # place nicely if you .connect() a lambda or other function/method
                # to PySignal

                # callback function for when the user changes the dropdown selection
                def setattr_func(value, key=None):
                    if key:
                        getattr(device, feat_name)[key] = keymapping_dict[widget.currentText()]
                    else:
                        setattr(device, feat_name, keymapping_dict[widget.currentText()])
                # call setattr_func() when the combo box value is set from the GUI
                widget.activated.connect(functools.partial(setattr_func, key=dictfeat_key))

                # callback function to modify the GUI when the the feat is changed externally
                def getattr_func(value, old_value, widget=widget):
                    # because the _changed is shared for all keys of the same dictfeat,
                    # we have to check to see if this was the key that was actually changed
                    widget.setCurrentIndex(list(keymapping_dict.values()).index(value))

        # the lantz feat is some sort of numerical value, so we will
        # generate a SpinBox (number entry with increment / decrement arrow keys)
        elif isinstance(feat_value, (int, float, Q_)):
            # arguments for the SpinBox
            optional_args = {}
            
            if feat._config['units']:
                if isinstance(feat_value, Q_):
                    base_units = infer_base_unit(feat_value)
                    base_units_str = '{0.units:~}'.format(Q_(1, base_units))
                    optional_args['suffix'] = base_units_str
                    optional_args['siPrefix'] = True
                else:
                    raise Exception('Didn\'t think this could happen... '
                        'the value of a lantz feat with units isn\'t a Q_')

            if feat._config['limits']:
                if len(feat._config['limits']) == 1:
                    # only min or only max was specified e.g.
                    # (,max) or (min,)
                    try:
                        optional_args['min'] = feat._config['limits'][0]
                    except IndexError:
                        optional_args['max'] = feat._config['limits'][1]
                else:
                    # (min, max) was specified
                    optional_args['bounds'] = feat._config['limits']

            # stepping strategy
            optional_args['dec'] = False
            optional_args['minStep'] = 1e-6
            # number of decimal places to display
            optional_args['decimals'] = 6
            optional_args['compactHeight'] = False
            if isinstance(feat_value, int):
                optional_args['int'] = True
                optional_args['minStep'] = 1
                # optional_args['decimals'] = 10
            
            widget = pyqtgraph_SpinBox(**optional_args)

            widget.resize(79, 24)
            # print(widget.sizeHint())
            def sizeHint(self):
                return QSize(79, 24)
            widget.sizeHint = sizeHint.__get__(widget, pyqtgraph_SpinBox)
            # print(widget.sizeHint())

            # widget.setSizeHint(QSize(120, 20))
            # print(widget.sizeHint())

            if isinstance(feat_value, Q_):
                widget.setValue(feat_value.to(feat._config['units']).m)
            else:
                widget.setValue(feat_value)
            
            if feat._config['units']:
                # callback function for when the user changes the value from the GUI
                def setattr_func(value):
                    setattr(device, feat_name, Q_(widget.value(), base_units))
                # callback function to modify the GUI when the the feat is changed externally
                def getattr_func(value, old_value, widget=widget):
                    if isinstance(value, Q_):
                        widget.setValue(value.to(base_units).m)
                    else:
                        widget.setValue(value)
            else:
                def setattr_func(value):
                    setattr(device, feat_name, widget.value())
                def getattr_func(value, old_value, widget=widget):
                    widget.setValue(value)
            
            if read_only:
                widget.lineEdit().setReadOnly(True)
            else:
                widget.sigValueChanged.connect(setattr_func)

        # the lantz feat is a string, so we will just make a text box
        elif isinstance(feat_value, str):
            widget = QLineEdit()
            widget.setText(feat_value)
            def setattr_func(value):
                setattr(device, feat_name, widget.text())

            if read_only:
                widget.setReadOnly(True)
            else:
                widget.textChanged.connect(setattr_func)

            def getattr_func(value, old_value, widget=widget):
                widget.setText(value)

        # some unknown type - make a readonly text box containing str(feat)
        else:
            widget = QLineEdit()
            widget.setText(str(feat))
            widget.setReadOnly(True)
            def getattr_func(value, old_value, widget=widget):
                widget.setText(value)

        return widget, getattr_func

    def _generate_action_widget(self, device, action, action_name):
        """Generate a Qt gui element for a lantz action"""
        action_button = QPushButton(action_name, self.tree)
        action_button.setFont(QFont('Helvetica [Cronyx]', 12))

        def action_func():
            getattr(device, action_name)()
        action_button.clicked.connect(action_func)

        return action_button

if __name__ ==  '__main__':
    import logging
    import sys
    from PyQt5.QtCore import Qt
    from nspyre.gui.app import NSpyreApp
    from nspyre.misc.logging import nspyre_init_logger

    nspyre_init_logger(logging.INFO)

    logger.info('starting Instrument Manager...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    pyqtgraph_connectCleanup()
    config_path = load_meta_config(CLIENT_META_CONFIG_PATH)
    with InservGateway(config_path) as isg:
        inserv_window = InstrumentManagerWindow(isg)
        app.exec()
        # inserv_window.close_connections()
    sys.exit()



#!/usr/bin/env python
"""
This GUI allows the running and managing of spyrelets

Author: Alexandre Bourassa
Date: 10/30/2019
"""

###########################
# imports
###########################

# std
import time
import traceback
import logging

# 3rd party
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtWidgets import QApplication, QMainWindow
from pyqtgraph import _connectCleanup as pyqtgraph_connectCleanup

# nspyre
from nspyre.config.config_files import load_meta_config
from nspyre.definitions import CLIENT_META_CONFIG_PATH
from nspyre.spyrelet.spyrelet import SpyreletLauncher
from nspyre.gui.widgets.param_widget import ParamWidget
from nspyre.gui.widgets.save_widget import Save_Widget
from nspyre.spyrelet.spyrelet import load_all_spyrelets
from nspyre.inserv.gateway import InservGateway

###########################
# globals
###########################

logger = logging.getLogger(__name__)

###########################
# classes
###########################

class Progress_Bar(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vars = []
        # self.iterators = []
        layout = QtWidgets.QVBoxLayout()
        self.pbar = QtWidgets.QProgressBar()
        self.pbar.setTextVisible(True)
        self.text = QtWidgets.QLabel('stopped')
        layout.addWidget(self.pbar)
        layout.addWidget(self.text)
        self.setLayout(layout)
    
    def reset(self):
        self.vars = []
        # self.iterators = []
        self.text.setText('stopped')

    def call_iter(self, iterable, max_val):
        # self.iterators.append(iter(iterable))
        self.vars.append({'val':0, 'max':max_val, 'start':time.time(), 'last':time.time(), 'avg':0, 'tot':0, 'rem':'?', 'per':0})

        if max_val != '?':
            self.pbar.setValue(0)
            self.pbar.setRange(0, max_val)

    def call_next(self):
        self.vars[-1]['val'] += 1
        t = time.time()
        self.vars[-1]['avg'] = t-self.vars[-1]['last']
        self.vars[-1]['last'] = t
        self.vars[-1]['tot'] = t-self.vars[-1]['start']

        if self.vars[-1]['max'] != '?':
            self.pbar.setValue(self.vars[-1]['val'])
            self.pbar.setRange(0, self.vars[-1]['max'])
            self.vars[-1]['per'] = 100*self.vars[-1]['val']/self.vars[-1]['max']
            self.vars[-1]['rem'] = (self.vars[-1]['max'] - self.vars[-1]['val'])*self.vars[-1]['avg']
        s_with_max = '[{per:.0f}% {val:.0f}/{max:.0f} [{tot:.0f}s<{rem:.0f}s] {avg:.2f}s/it]'
        s_without_max = '[{val:.0f}[{tot:.0f}s] {avg:.2f}s/it]'
        self.text.setText('\t'.join([(s_without_max if d['max']=='?' else s_with_max).format(**d)for d in self.vars]))
        QtWidgets.QApplication.processEvents()
    
    def call_stopiter(self):
        self.vars = self.vars[:-1]
        if not self.vars == [] and self.vars[-1]['max'] != '?':
            self.pbar.setValue(self.vars[-1]['max'])
            self.pbar.setRange(0, self.vars[-1]['max'])
            QtWidgets.QApplication.processEvents()
        # self.iterators = self.iterators[:-1]

    # def update(self, iterator):

class Spyrelet_Launcher_Widget(QtWidgets.QWidget):
    def __init__(self, spyrelet, parent=None):
        self.spyrelet = spyrelet
        self.progress_bar = Progress_Bar()
        self.launcher = SpyreletLauncher(spyrelet)
        self.param_w = ParamWidget(self.launcher.params)
        self.param_w.set(**self.launcher.get_defaults())
        super().__init__(parent=parent)

        #Build ctrl pannel
        ctrl_pannel = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout()
        self.run_btn = QtWidgets.QPushButton('Run')
        self.set_defaults_btn = QtWidgets.QPushButton('Set Defaults')
        self.save_btn = QtWidgets.QPushButton('Save')
        layout.addWidget(self.run_btn)
        layout.addWidget(self.save_btn)
        ctrl_pannel.setLayout(layout)

        #Build main layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(ctrl_pannel)
        layout.addWidget(self.progress_bar)
        layout.addWidget(self.param_w)
        layout.addStretch()
        self.setLayout(layout)

        #Connect signal
        self.run_btn.clicked.connect(self.run)
        self.set_defaults_btn.clicked.connect(self.set_defaults)
        self.save_btn.clicked.connect(self.save_data)

        self.run_thread = None
        

    def run(self):
        if self.run_thread is None or self.run_thread.isFinished():
            self.progress_bar.reset()
            param_dict = self.param_w.get()
            self.run_thread = Spyrelet_Run_Thread(self.launcher, param_dict)
            self.run_thread.start()
            self.run_btn.setText('Stop')
            self.run_thread.finished.connect(lambda: self.run_btn.setText('Run'))
            self.run_thread.progressed_iter.connect(self.progress_bar.call_iter)
            self.run_thread.progressed_next.connect(self.progress_bar.call_next)
            self.run_thread.progressed_stopiter.connect(self.progress_bar.call_stopiter)
        else:
            self.run_thread.stop_requested.emit()

    def set_defaults(self):
        self.spyrelet.set_defaults(**self.param_w.get())

    def save_data(self):
        self.save_w = Save_Widget(self.spyrelet)
        self.save_w.show()

class Spyrelet_Run_Thread(QtCore.QThread):
    """Qt Thread which monitors for changes to qither a collection or a database and emits a signal when something happens"""
    progressed_iter = QtCore.pyqtSignal(object, object) #This will be emitted when iter is called on progress is called in the spyrelet
    progressed_next = QtCore.pyqtSignal() #This will be emitted when next is called on progress is called in the spyrelet
    progressed_stopiter = QtCore.pyqtSignal() #This will be emitted when next is called on progress is called in the spyrelet
    stop_requested = QtCore.pyqtSignal() # This will be emitted externally to stop the execution of a spyrelet prematurly
    def __init__(self, launcher, param_dict):
        super().__init__()
        self.param_dict = param_dict
        self.launcher = launcher
        self.spyrelet = self.launcher.spyrelet
        class Progress_Iter():
            def __init__(_self, iterable):
                _self.iterable = iterable
            def __iter__(_self):
                try:
                    max_val = len(_self.iterable)
                except TypeError:
                    traceback.print_exc()
                    max_val = '?'
                _self.iterable = iter(_self.iterable)
                self.progressed_iter.emit(_self.iterable, max_val)
                return _self
            def __next__(_self):
                try:
                    val = next(_self.iterable)
                    self.progressed_next.emit()
                except StopIteration as e:
                    self.progressed_stopiter.emit()
                    raise e
                return val


        self.progress = Progress_Iter
        self.stop_requested.connect(self.stop_run)

    def progress(self, iterator):
        for x in iterator:
            self.progressed.emit(iterator)
            yield x

    def run(self):
        self.launcher.run(progress=self.progress, **self.param_dict)

    def stop_run(self):
        self.spyrelet.stop()


class CombinedSpyreletWindow(QMainWindow):
    def __init__(self, gateway, spyrelets=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('NSpyre Instrument Manager')

        # Set the main window layout to consist of vertical boxes.
        # The QVBoxLayout class lines up widgets vertically.
        layout = QtWidgets.QVBoxLayout()

        self.selector = QtWidgets.QComboBox()
        container = QtWidgets.QWidget()
        layout.addWidget(self.selector)
        layout.addWidget(container)
        #self.setLayout(layout)

        #Create the launchers
        spyrelets = load_all_spyrelets(gateway) if spyrelets is None else spyrelets
        self.launchers = {name: Spyrelet_Launcher_Widget(s) for name, s in spyrelets.items()}
        # cfg = get_configs()
        # names = list(cfg['experiment_list'].keys())
        # names.sort(key=lambda x: len(cfg['experiment_list'][x][2]))
        # self.launchers = dict()
        # last_len = -1
        # while last_len != len(self.launchers):
        #     last_len = len(self.launchers)
        #     for sname in names:
        #         sclass, devs, subs = cfg['experiment_list'][sname]
        #         print(sname, all([x in self.launchers for x in list(subs.values())]))
        #         if not sname in self.launchers and all([x in self.launchers for x in list(subs.values())]):
        #             try:
        #                 sclass = get_class_from_str(sclass)
        #                 subs = {real_name:self.launchers[alias].spyrelet for real_name,alias in subs.items()}
        #                 s = sclass(sname, spyrelets=subs, device_alias=devs)
        #                 self.launchers[sname] = Spyrelet_Launcher_Widget(s)
        #             except:
        #                 print("Could not instanciate launcher for spyrelet {}...".format(sname))
        #                 traceback.print_exc()

        #Add to layout
        stacked_layout = QtWidgets.QStackedLayout()
        names = list(self.launchers.keys())
        names.sort()
        for n in names:
            stacked_layout.addWidget(self.launchers[n])
        container.setLayout(stacked_layout)

        self.selector.addItems(names)
        self.container_layout = stacked_layout
        self.selector.currentTextChanged.connect(self.change_widget)
        w = QtWidgets.QWidget()
        w.setLayout(layout)
        self.setCentralWidget(w)
        self.show()

    def change_widget(self, name):
        self.container_layout.setCurrentWidget(self.launchers[name])


if __name__ ==  '__main__':
    import logging
    import sys
    from PyQt5.QtCore import Qt
    from nspyre.gui.app import NSpyreApp
    from nspyre.misc.logging import nspyre_init_logger

    nspyre_init_logger(logging.INFO)

    logger.info('starting Instrument Manager...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    pyqtgraph_connectCleanup()
    config_path = load_meta_config(CLIENT_META_CONFIG_PATH)
    with InservGateway(config_path) as isg:
        combined_spyrelet_window = CombinedSpyreletWindow(isg)
        sys.exit(app.exec())



#!/usr/bin/env python
"""
    nspyre.gui.__init__.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    This module runs the main nspyre GUI, which allows the user to easily
    launch other tools

    Author: Alexandre Bourassa
    Modified: Jacob Feder 7/25/2020, Michael Solomon 9/12/2020
"""

import logging
import sys
from pathlib import Path

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication

from .app import NSpyreApp
from .main_window import NSpyreMainWindow

from nspyre.misc.logging import nspyre_init_logger

###########################
# globals
###########################

logger = logging.getLogger(__name__)

###########################
# standalone main
###########################

def main(args=None):
    """Entry point for the application script"""

    nspyre_init_logger(logging.INFO)

    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp(['NSpyre', sys.argv])
    main_window = NSpyreMainWindow()
    sys.exit(app.exec())



import nspyre.spyrelet.spyrelet
from nspyre.inserv.gateway import InservGateway

with InservGateway() as im:
    sp = nspyre.spyrelet.spyrelet.load_all_spyrelets(im)
    nspyre.spyrelet.spyrelet.unload_all_spyrelets()
    sp = nspyre.spyrelet.spyrelet.load_all_spyrelets(im)
    nspyre.spyrelet.spyrelet.unload_spyrelet('s2')
    try:
        sp = nspyre.spyrelet.spyrelet.load_all_spyrelets(im)
    except:
        pass
    nspyre.spyrelet.spyrelet.load_spyrelet('s2', im)
    nspyre.spyrelet.spyrelet.reload_all_spyrelets(im)
    nspyre.spyrelet.spyrelet.reload_spyrelet('s2', im)



#!/usr/bin/env python
"""
    ?

    Author: Alexandre Bourassa
"""

###########################
# imports
###########################

# std
import time
import inspect
import traceback
import textwrap
import logging

# 3rd party
from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QMainWindow
import numpy as np

# nspyre
from nspyre.gui.widgets.plotting import LinePlotWidget, HeatmapPlotWidget
from nspyre.gui.widgets.splitter_widget import Splitter, SplitterOrientation
from nspyre.mongodb.mongo_listener import Synched_Mongo_Database
from nspyre.gui.widgets.views import Spyrelet_Views
from nspyre.definitions import join_nspyre_path
from nspyre.spyrelet.spyrelet import custom_decode
from nspyre.misc.misc import cleanup_register
from nspyre.gui.image import ImageWidget
from nspyre.gui.widgets.code_editor import Scintilla_Code_Editor, \
                Monokai_Python_Lexer

###########################
# globals
###########################

logger = logging.getLogger(__name__)

###########################
# classes
###########################

class CustomView():
    def __init__(self, code_editor, w1D, w2D, plot_layout, initial_df, initial_cache):
        self.is_updating = False
        self.editor = code_editor
        self.plot_layout = plot_layout
        self.plot_type = '1D'
        self.w = w1D
        self.ws = {'1D':w1D, '2D':w2D}
        self.valid_code = False
        self._source = "plot_type = '1D'\ndef plot(df, cache):\n    return {}"
        self.last_df = initial_df
        self.last_cache = initial_cache
        self.editor.request_run_signal.connect(self.analyze)
        
    
    def start_updating(self):
        self.analyze()
        self.w.clear()
        if self.valid_code:
            self.is_updating = True
    
    def stop_updating(self):
        self.is_updating = False

    def get_source(self):
        return self._source

    def analyze(self):
        self.valid_code = False
        source = self.editor.text()
        self._source = source
        try:
            exec(source)
        except:
            print("Error in executing the code")
            traceback.print_exc()
            return
        if not ('plot_type' in locals() and 'plot' in locals()):
            print("Missing either a 'plot_type' variable or a 'plot' function")
            return
        _plot_type, plot_fun = locals()['plot_type'], locals()['plot']
        if _plot_type in ['1D', '2D']:
            self.plot_type = _plot_type
            self.w = self.ws[_plot_type]
            self.plot_layout.setCurrentWidget(self.w)
        else:
            print("Invalid plot type!  Must be either '1D' or '2D'")
            return
        
        self.update_fun = plot_fun
        self.valid_code = True
        self.update(self.last_df, self.last_cache)

    def update(self, df, cache):
        self.last_df = df
        self.last_cache = cache
        if self.is_updating and self.valid_code:
            if self.plot_type == '1D':
                traces = self.update_fun(df, cache)
                for name, data in traces.items():
                    self.w.set(name, xs=data[0], ys=data[1])
            elif self.plot_type == '2D':
                im = np.array(self.update_fun(df, cache))
                self.w.set(im)

class BaseView():
    def __init__(self, view, w):
        self.view = view
        self.is_updating = False
        self.update_fun = view.update_fun
        self.init_formatter = view.get_formatter(self.w, 'init')
        self.update_formatter = view.get_formatter(self.w, 'update')
        # if not self.init_formatter is None:
        #     self.init_formatter(self.w)
            
    def start_updating(self):
        self.w.clear()
        if not self.init_formatter is None:
            self.init_formatter(self.w)
        self.is_updating = True

    def stop_updating(self):
        self.is_updating = False

    # def update(self, df):
    #     if self.is_updating:
    #         traces = self.update_fun(df)
    #         for name, data in traces.items():
    #             self.w.set(name, xs=data[0], ys=data[1])
    #         if not self.update_formatter is None:
    #             self.update_formatter(self.w, df)

    def get_source(self):
        return textwrap.dedent(inspect.getsource(self.update_fun))

class LinePlotView(BaseView):
    def __init__(self, view, w=None):
        if view.type != '1D':
            raise "This class is for 1D plot only"
        if w is None:
            self.w = LinePlotWidget()
            self.w.setFont(QFont('Helvetica [Cronyx]', 12))
        else:
            self.w = w
        super().__init__(view, self.w)
            
    def update(self, df, cache):
        if self.is_updating:
            traces = self.update_fun(df, cache)
            for name, data in traces.items():
                self.w.set(name, xs=data[0], ys=data[1])
            if not self.update_formatter is None:
                self.update_formatter(self.w, df, cache)

class HeatmapPlotView(BaseView):
    def __init__(self, view, w=None):
        if view.type != '2D':
            raise "This class is for 2D plot only"
        if w is None:
            self.w = HeatmapPlotWidget()
            self.w.setFont(QFont('Helvetica [Cronyx]', 12))
        else:
            self.w = w
        super().__init__(view, self.w)
            
    def update(self, df, cache):
        if self.is_updating:
            im = np.array(self.update_fun(df, cache))
            self.w.set(im)
            if not self.update_formatter is None:
                self.update_formatter(self.w, df, cache)


class ViewManagerWindow(QMainWindow):
    def __init__(self, mongodb_addr=None, db_name='Spyre_Live_Data', react_to_drop=False, db=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('NSpyre View Manager')
        self.setFont(QFont('Helvetica [Cronyx]'))
        if db is None:
            self.db = Synched_Mongo_Database(db_name, mongodb_addr=mongodb_addr)
        else:
            self.db = db
        # cleanup_register(mongodb_addr)

        self.views = dict()
        self.items = dict()
        self.last_updated = dict()
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update_colors)
        self.timer.start(100)
        self.fade_rate = 2
        
        layout = QtWidgets.QHBoxLayout()
        # Build tree
        self.tree = QtWidgets.QTreeWidget()
        self.tree.setColumnCount(1)
        self.tree.setHeaderHidden(True)

        layout.addWidget(self.tree)

        # Build layout
        self.plot_layout = QtWidgets.QStackedLayout()
        plot_container = QtWidgets.QWidget()
        plot_container.setLayout(self.plot_layout)
        self.default_image = ImageWidget(str(\
                                join_nspyre_path('images/logo.png')))
        self.plot_layout.addWidget(self.default_image)
        
        self.common_lineplotwidget = LinePlotWidget()
        self.common_heatmapplotwidget = HeatmapPlotWidget()
        self.plot_layout.addWidget(self.common_lineplotwidget)
        self.plot_layout.addWidget(self.common_heatmapplotwidget)


        splitter_config = {
            'main_w': self.tree,
            'side_w': plot_container,
            'orientation': SplitterOrientation.vertical_right_button,
        }
        hsplitter = Splitter(**splitter_config)
        hsplitter.setSizes([1, 400])
        hsplitter.setHandleWidth(10)

        self.code_editor = Scintilla_Code_Editor()
        lexer = Monokai_Python_Lexer(self.code_editor)
        self.code_editor.setLexer(lexer)
        splitter_config = {
            'main_w': hsplitter,
            'side_w': self.code_editor,
            'orientation': SplitterOrientation.horizontal_top_button,
        }
        vsplitter = Splitter(**splitter_config)
        vsplitter.setSizes([1, 400])
        vsplitter.setHandleWidth(10)

        #layout = QtWidgets.QHBoxLayout()
        #layout.addWidget(vsplitter)
        #self.setLayout(layout)
        #w = QWidget()
        #w.setLayout(layout)
        self.setCentralWidget(vsplitter)

        if 'Register' in self.db.dfs:
            names = list(self.db.get_df('Register').index)
            names.sort()
            for name in names:
                self.add_col(name, try_update=False)


        self.tree.currentItemChanged.connect(self.new_table_selection)

        #Connect db signals
        self.db.col_added.connect(self.add_col)
        self.db.updated_row.connect(self._update_plot)
        if react_to_drop:
            self.db.col_dropped.connect(self.del_col)
        self.show()

    def _update_plot(self, col_name, row):
        if col_name != 'Register':
            self.last_updated[col_name] = time.time()
        self.update_plot(col_name, row)

    def update_plot(self, col_name, row):
        if col_name == 'Register':
            self.add_col(row.name)
            return
        for name, view in self.views[col_name].items():
            if col_name in self.db.get_df('Register').index and 'cache' in self.db.get_df('Register').loc[col_name] and not self.db.get_df('Register').loc[col_name]['cache'] is np.nan:
                cache = custom_decode(self.db.get_df('Register').loc[col_name]['cache'])
            else:
                cache = {}
            view.update(self.db.get_df(col_name), cache)
        
        
    def add_col(self, col_name, try_update=True):
        if col_name == 'Register':
            return
        sclass = self.db.get_df('Register').loc[col_name].name
        spyrelet_views = Spyrelet_Views(sclass)
        
        if col_name in self.views:
            return
        top = QtWidgets.QTreeWidgetItem(self.tree, [col_name])
        self.default_color = top.background(0) # This is used to remember the default color when instanciating (to restore in update_color)

        self.items[col_name] = {'__top':top}
        self.views[col_name] = dict()
        self.last_updated[col_name] = time.time()

        for name, view in spyrelet_views.views.items():
            if view.type == '1D':
                self.views[col_name][name] = LinePlotView(view, self.common_lineplotwidget)
            elif view.type == '2D':
                self.views[col_name][name] = HeatmapPlotView(view, self.common_heatmapplotwidget)
            self.items[col_name][name] = QtWidgets.QTreeWidgetItem(0)#QtGui.QStandardItem(name)
            self.items[col_name][name].setText(0, name)
            
            # self.plot_layout.addWidget(self.views[col_name][name].w)
            top.addChild(self.items[col_name][name])


        self.tree.insertTopLevelItem(0, top)
        try:
            self.update_plot(col_name, None)
        except:
            pass

    def make_new_custom_view(self, col_name):
        name_template = 'Custom {}'
        for i in range(100):
            name = name_template.format(i)
            if not name in self.views[col_name]:
                if col_name in self.db.get_df('Register').index and 'cache' in self.db.get_df('Register').loc[col_name] and not self.db.get_df('Register').loc[col_name]['cache'] is np.nan:
                    cache = custom_decode(self.db.get_df('Register').loc[col_name]['cache'])
                else:
                    cache = {}
                self.views[col_name][name] = CustomView(self.code_editor, self.common_lineplotwidget, self.common_heatmapplotwidget, self.plot_layout, self.db.get_df(col_name), cache)
                self.items[col_name][name] = QtWidgets.QTreeWidgetItem(0)
                self.items[col_name][name].setText(0, name)
                self.items[col_name]['__top'].addChild(self.items[col_name][name])
                break


    def del_col(self, col_name):
        if col_name == 'Register':
            return
        # for pname, view in self.views[col_name].items():
        #     self.plot_layout.removeWidget(view.w)
        #     view.w.deleteLater()
        for iname, item in self.items[col_name].items():
            self.tree.removeItemWidget(item,0)
        self.tree.takeTopLevelItem(self.tree.indexOfTopLevelItem(self.items[col_name]['__top']))
        self.views.pop(col_name)
        self.items.pop(col_name)
        self.last_updated.pop(col_name)
    
    def get_view_name(self, item):
            if item is None:
                return None, None
            txt = item.text(0)
            if not txt in self.views:
                spyrelet_name = item.parent().text(0)
                view_name = txt
                return spyrelet_name, view_name
            else:
                # Selected a top level item
                return txt, None

    def new_table_selection(self, cur, prev):
        spyrelet_name, view_name = self.get_view_name(cur)
        spyrelet_name_old, view_name_old = self.get_view_name(prev)
        if not view_name_old is None:
            self.views[spyrelet_name_old][view_name_old].stop_updating()
        if view_name is None:
            # Selected a top level item
            self.plot_layout.setCurrentWidget(self.default_image)
            return
        
        self.code_editor.setText(self.views[spyrelet_name][view_name].get_source())
        self.plot_layout.setCurrentWidget(self.views[spyrelet_name][view_name].w)
        self.views[spyrelet_name][view_name].start_updating()
        if spyrelet_name in self.db.dfs:
            self.update_plot(spyrelet_name, None)

    def update_colors(self):
        try:
            for col_name, last_time in self.last_updated.items():
                delta = (time.time()-last_time)/self.fade_rate
                if delta<1:
                    color = QtGui.QColor(0, 255, 0, int(max((1-delta)*255,0)))
                    self.items[col_name]['__top'].setBackground(0, QtGui.QBrush(color))
                else:
                    self.items[col_name]['__top'].setBackground(0, self.default_color)
        except:
            pass

    def keyPressEvent(self, ev):
        if type(ev)==QtGui.QKeyEvent:
            sname, vname = self.get_view_name(self.tree.currentItem())
            if ev.matches(QtGui.QKeySequence.Copy):
                # app = QtWidgets.QApplication.instance()
                # app.clipboard().setText(sname)
                self.make_new_custom_view(sname)
                ev.accept()
            if ev.matches(QtGui.QKeySequence.Save):
                # filename, other = QtWidgets.QFileDialog.getSaveFileName(None, 'Save this table to...', '', 'Comma Separated Value (*.csv)')
                # if filename:
                print(self.db.get_df(sname))
                    # df = self.df_table.model()._data
                    # df.to_csv(path_or_buf=filename, sep=',')
                ev.accept()


if __name__ ==  '__main__':
    import logging
    import sys
    from PyQt5.QtCore import Qt
    from nspyre.gui.app import NSpyreApp
    from nspyre.misc.logging import nspyre_init_logger

    nspyre_init_logger(logging.INFO)

    logger.info('starting View Manager...')
    if hasattr(Qt, 'AA_EnableHighDpiScaling'):
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

    if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
    app = NSpyreApp([sys.argv])
    data_view_window = ViewManagerWindow(react_to_drop=False)
    sys.exit(app.exec())
